"""
Datastructure(s) for use in settings.py

Note: do not place any Django-specific imports in this file, as
it must be imported in settings.py.
"""

from dataclasses import dataclass
from typing import Union


@dataclass
class ContentType:
    """
    Data class for keeping track of content types and associated default encodings
    """

    pattern: str
    default_encoding: str


@dataclass
class ProcessedBody:
    allow_saving_to_db: bool
    content: Union[bytes, str]
    content_type: str
    encoding: str



import logging

from django.utils import timezone

from requests import Session

logger = logging.getLogger("requests")


def hook_requests_logging(response, *args, **kwargs):
    """
    A hook for requests library in order to add extra data to the logs
    """
    extra = {"requested_at": timezone.now(), "req": response.request, "res": response}
    logger.debug("Outgoing request", extra=extra)


def install_outgoing_requests_logging():
    """
    Log all outgoing requests which are made by the library requests during a session.
    """

    if hasattr(Session, "_initial_request"):
        logger.debug(
            "Session is already patched OR has an ``_initial_request`` attribute."
        )
        return

    Session._initial_request = Session.request  # type: ignore

    def new_request(self, *args, **kwargs):
        self.hooks["response"].append(hook_requests_logging)
        return self._initial_request(*args, **kwargs)

    Session.request = new_request



import logging
import textwrap

from django.conf import settings


class HttpFormatter(logging.Formatter):
    def _formatHeaders(self, d):
        return "\n".join(f"{k}: {v}" for k, v in d.items())

    def _formatBody(self, content: str, request_or_response: str) -> str:
        if settings.LOG_OUTGOING_REQUESTS_EMIT_BODY:
            return f"\n{request_or_response} body:\n{content}"
        return ""

    def formatMessage(self, record):
        result = super().formatMessage(record)

        if record.name != "requests":
            return result

        result += textwrap.dedent(
            """
            ---------------- request ----------------
            {req.method} {req.url}
            {reqhdrs} {request_body}

            ---------------- response ----------------
            {res.status_code} {res.reason} {res.url}
            {reshdrs} {response_body}

        """
        ).format(
            req=record.req,
            res=record.res,
            reqhdrs=self._formatHeaders(record.req.headers),
            reshdrs=self._formatHeaders(record.res.headers),
            request_body=self._formatBody(record.req.body, "Request"),
            response_body=self._formatBody(record.res.content, "Response"),
        )

        return result



"""Tests for the utility functions"""

import logging

import pytest
import requests

from log_outgoing_requests.datastructures import ContentType
from log_outgoing_requests.models import OutgoingRequestsLogConfig
from log_outgoing_requests.utils import (
    check_content_length,
    check_content_type,
    get_default_encoding,
    parse_content_type_header,
)


#
# test check_content_length
#
@pytest.mark.parametrize(
    "max_content_length, expected",
    [
        (1048, True),
        (12, False),
    ],
)
@pytest.mark.django_db
def test_check_content_length(
    max_content_length,
    requests_mock,
    request_mock_kwargs,
    expected,
):
    config = OutgoingRequestsLogConfig.objects.create(
        max_content_length=max_content_length,
    )

    # we check the functionality of determining missing content length
    # only for the response (the requests library automatically inserts
    # the content length for the request)
    del request_mock_kwargs["headers"]["Content-Length"]

    mock = requests_mock.post(**request_mock_kwargs)
    response = requests.post(
        request_mock_kwargs["url"],
        headers=request_mock_kwargs["request_headers"],
        json={"test": "request data"},
    )

    assert response.status_code == 200

    result_request = check_content_length(mock.last_request, config=config)
    assert result_request is expected

    result_response = check_content_length(response, config=config)
    assert result_response is expected


#
# test parse_content_type_header
#
@pytest.mark.parametrize(
    "content_type_header, expected_content_type, expected_encoding",
    [
        ("text/xml; charset=us-ascii", "text/xml", "us-ascii"),
        ("text/xml", "text/xml", ""),
        ("application/json", "application/json", ""),
        ("", "", ""),
    ],
)
@pytest.mark.django_db
def test_parse_content_type_header(
    content_type_header,
    requests_mock,
    request_mock_kwargs,
    expected_content_type,
    expected_encoding,
):
    request_mock_kwargs["request_headers"]["Content-Type"] = content_type_header
    request_mock_kwargs["headers"]["Content-Type"] = content_type_header

    mock = requests_mock.post(**request_mock_kwargs)
    response = requests.post(
        request_mock_kwargs["url"],
        headers=request_mock_kwargs["request_headers"],
        json={"test": "request data"},
    )

    assert response.status_code == 200

    # check request
    parsed_request_header = parse_content_type_header(mock.last_request)
    assert parsed_request_header[0] == expected_content_type
    assert parsed_request_header[1] == expected_encoding

    # check response
    parsed_response_header = parse_content_type_header(response)
    assert parsed_response_header[0] == expected_content_type
    assert parsed_response_header[1] == expected_encoding


#
# test check_content_type
#
@pytest.mark.parametrize(
    "allowed_content_types, content_type_pattern, expected",
    [
        ([ContentType("text/xml", "iso-8859-1")], "text/xml", True),
        ([ContentType("text/xml", "iso-8859-1")], "text/html", False),
        ([ContentType("text/xml", "iso-8859-1")], "video/mp4", False),
        ([ContentType("text/*", "utf-8")], "text/html", True),
    ],
)
@pytest.mark.django_db
def test_check_content_type(
    allowed_content_types,
    content_type_pattern,
    expected,
    settings,
):
    settings.LOG_OUTGOING_REQUESTS_CONTENT_TYPES = allowed_content_types

    result = check_content_type(content_type_pattern)
    assert result is expected


#
# test get_default_encoding
#
@pytest.mark.parametrize(
    "content_type_pattern, expected",
    [
        ("text/html", "utf-8"),
        ("text/xml", "iso-8859-1"),
        ("application/json", "utf-8"),
        ("application/unknown", ""),
    ],
)
@pytest.mark.django_db
def test_get_default_encoding(
    content_type_pattern,
    expected,
):
    result = get_default_encoding(content_type_pattern)
    assert result == expected


@pytest.mark.django_db
def test_logger_warning_missing_content_length(
    requests_mock, request_mock_kwargs, caplog
):
    del request_mock_kwargs["request_headers"]["Content-Length"]

    with caplog.at_level(logging.DEBUG):
        requests_mock.get(**request_mock_kwargs)
        requests.get(
            request_mock_kwargs["url"], headers=request_mock_kwargs["request_headers"]
        )

    records = caplog.records
    assert records[1].levelname == "WARNING"
    assert records[1].name == "log_outgoing_requests.utils"
    assert (
        records[1].msg
        == "Content length of the request/response (request netloc: example.com:8000) could not be determined."
    )



from django.db import models
from django.utils.translation import gettext_lazy as _


class SaveLogsChoice(models.TextChoices):
    use_default = "use_default", _("Use default")
    yes = "yes", _("Yes")
    no = "no", _("No")



"""Tests for the HttpFormatter helper class"""

import logging

import pytest
import requests

from log_outgoing_requests.formatters import HttpFormatter


@pytest.mark.django_db
@pytest.mark.parametrize(
    "log_body, expected",
    [
        (True, True),
        (False, False),
    ],
)
def test_formatter(
    requests_mock,
    request_mock_kwargs,
    caplog,
    settings,
    log_body,
    expected,
):
    """Assert that request/response bodies are (not) saved if setting is enabled (disabled)"""

    settings.LOG_OUTGOING_REQUESTS_EMIT_BODY = log_body

    formatter = HttpFormatter()

    with caplog.at_level(logging.DEBUG):
        requests_mock.post(**request_mock_kwargs)
        requests.post(
            request_mock_kwargs["url"],
            headers=request_mock_kwargs["request_headers"],
            json={"test": "request data"},
        )

    record = caplog.records[1]

    res = formatter.formatMessage(record)

    assert ('{"test": "request data"}' in res) is expected
    assert ('{"test": "response data"}' in res) is expected



"""Tests for the utility functions"""

import logging
from typing import Iterable, Tuple, Union

from django.conf import settings

from requests import PreparedRequest, Response

from .compat import parse_header_parameters
from .datastructures import ContentType, ProcessedBody
from .models import OutgoingRequestsLogConfig

logger = logging.getLogger(__name__)

HttpObj = Union[PreparedRequest, Response]


def process_body(http_obj: HttpObj, config: OutgoingRequestsLogConfig) -> ProcessedBody:
    """
    Process a request or response body by parsing the meta information.
    """
    content_type, encoding = parse_content_type_header(http_obj)
    if not encoding:
        encoding = get_default_encoding(content_type)
    allow_persisting = check_content_type(content_type) and check_content_length(
        http_obj, config=config
    )
    content = _get_body(http_obj) if allow_persisting else b""
    return ProcessedBody(
        allow_saving_to_db=allow_persisting,
        content=content or b"",
        content_type=content_type,
        encoding=encoding,
    )


#
# Handler utilities
#
def _get_body(http_obj: HttpObj) -> Union[bytes, str, None]:
    return http_obj.content if isinstance(http_obj, Response) else http_obj.body


def _get_content_length(http_obj: HttpObj) -> str:
    """
    Try to determine the size of a request/response content.

    Try `Content-Length` header first. If not present, try to
    determine the size by reading `len(body)`. The entire content
    is thereby read into memory (the assumption being that the content
    will eventually be consumed anyway).
    """
    content_length = http_obj.headers.get("Content-Length", "")

    if not content_length:
        body = _get_body(http_obj)
        if body is not None:
            content_length = str(len(body))

    return content_length


def check_content_length(
    http_obj: HttpObj,
    config: "OutgoingRequestsLogConfig",
) -> bool:
    """
    Check `content_length` against settings.

    If `content_length` could not be determined (i.e. `content_length` == ""), the test
    passes with a warning.
    """
    content_length = _get_content_length(http_obj)

    if not content_length:
        # for logging: get netloc (response) or url (request)
        target = getattr(http_obj, "netloc", "") or http_obj.url
        logger.warning(
            "Content length of the request/response (request netloc: %s) could not be determined."
            % target
        )
        return True

    max_content_length = config.max_content_length

    return int(content_length) <= max_content_length


def parse_content_type_header(http_obj: HttpObj) -> Tuple[str, str]:
    """
    Wrapper around Django's `parse_header`.

    If a charset/encoding is found, we replace the representation of it with a string.

    :returns: a `tuple` (content_type, encoding)
    """
    content_type_line = http_obj.headers.get("Content-Type", "")
    if not content_type_line:
        return ("", "")

    content_type, params = parse_header_parameters(content_type_line)
    encoding = params.get("charset", "")
    return content_type, encoding


def check_content_type(content_type: str) -> bool:
    """
    Check `content_type` against settings.

    The string patterns of the content types specified under `LOG_OUTGOING_REQUESTS_
    CONTENT_TYPES` are split into two groups. For regular patterns not containing a
    wildcard ("text/xml"), check if `content_type.pattern` is included in the list.
    For patterns containing a wildcard ("text/*"), check if `content_type.pattern`
    is a substring of any pattern contained in the list.
    """
    allowed_content_types: Iterable[
        ContentType
    ] = settings.LOG_OUTGOING_REQUESTS_CONTENT_TYPES
    regular_patterns = [
        item.pattern for item in allowed_content_types if not item.pattern.endswith("*")
    ]
    wildcard_patterns = [
        item.pattern for item in allowed_content_types if item.pattern.endswith("*")
    ]

    if content_type in regular_patterns:
        return True

    return any(content_type.startswith(pattern[:-1]) for pattern in wildcard_patterns)


def get_default_encoding(content_type_pattern: str) -> str:
    """
    Get the default encoding for the `ContentType` with the associated pattern.
    """
    allowed_content_types: Iterable[
        ContentType
    ] = settings.LOG_OUTGOING_REQUESTS_CONTENT_TYPES

    regular_types = [
        item for item in allowed_content_types if not item.pattern.endswith("*")
    ]
    wildcard_types = [
        item for item in allowed_content_types if item.pattern.endswith("*")
    ]

    content_type = next(
        (item for item in regular_types if item.pattern == content_type_pattern), None
    )
    if content_type is None:
        content_type = next(
            (
                item
                for item in wildcard_types
                if content_type_pattern.startswith(item.pattern[:-1])
            ),
            None,
        )

    return content_type.default_encoding if content_type else ""



from django.apps import AppConfig


class LogOutgoingRequestsConfig(AppConfig):
    name = "log_outgoing_requests"

    def ready(self):
        from .log_requests import install_outgoing_requests_logging

        install_outgoing_requests_logging()



# Generated by Django 3.2.1 on 2023-06-02 16:22

import django.core.validators
from django.db import migrations, models

import log_outgoing_requests.models


class Migration(migrations.Migration):
    dependencies = [
        ("log_outgoing_requests", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="OutgoingRequestsLogConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "save_to_db",
                    models.CharField(
                        choices=[
                            ("use_default", "Use default"),
                            ("yes", "Yes"),
                            ("no", "No"),
                        ],
                        default="use_default",
                        max_length=11,
                        verbose_name="Save logs to database",
                    ),
                ),
                (
                    "save_body",
                    models.CharField(
                        choices=[
                            ("use_default", "Use default"),
                            ("yes", "Yes"),
                            ("no", "No"),
                        ],
                        default="use_default",
                        max_length=11,
                        verbose_name="Save request + response body",
                    ),
                ),
                (
                    "max_content_length",
                    models.IntegerField(
                        default=log_outgoing_requests.models.get_default_max_content_length,
                        help_text="The maximal size of the request/response content (in bytes). If 'Require content length' is not checked, this setting has no effect.",
                        validators=[django.core.validators.MinValueValidator(0)],
                        verbose_name="Maximal content size",
                    ),
                ),
            ],
            options={
                "verbose_name": "Outgoing Requests Log Configuration",
            },
        ),
        migrations.AddField(
            model_name="outgoingrequestslog",
            name="req_body",
            field=models.BinaryField(
                default=b"", help_text="The request body.", verbose_name="Request body"
            ),
        ),
        migrations.AddField(
            model_name="outgoingrequestslog",
            name="req_body_encoding",
            field=models.CharField(default="", max_length=24),
        ),
        migrations.AddField(
            model_name="outgoingrequestslog",
            name="res_body",
            field=models.BinaryField(
                default=b"",
                help_text="The response body.",
                verbose_name="Response body",
            ),
        ),
        migrations.AddField(
            model_name="outgoingrequestslog",
            name="res_body_encoding",
            field=models.CharField(default="", max_length=24),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="hostname",
            field=models.CharField(
                default="",
                help_text="The netloc/hostname part of the url.",
                max_length=255,
                verbose_name="Hostname",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="method",
            field=models.CharField(
                blank=True,
                help_text="The type of request method.",
                max_length=10,
                verbose_name="Method",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="req_content_type",
            field=models.CharField(
                default="",
                help_text="The content type of the request.",
                max_length=50,
                verbose_name="Request content type",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="req_headers",
            field=models.TextField(
                default="",
                help_text="The request headers.",
                verbose_name="Request headers",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="res_content_type",
            field=models.CharField(
                default="",
                help_text="The content type of the response.",
                max_length=50,
                verbose_name="Response content type",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="res_headers",
            field=models.TextField(
                default="",
                help_text="The response headers.",
                verbose_name="Response headers",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="response_ms",
            field=models.PositiveIntegerField(
                default=0,
                help_text="This is the response time in ms.",
                verbose_name="Response in ms",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="trace",
            field=models.TextField(
                default="",
                help_text="Text providing information in case of request failure.",
                verbose_name="Trace",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingrequestslog",
            name="url",
            field=models.URLField(
                default="",
                help_text="The url of the outgoing request.",
                max_length=1000,
                verbose_name="URL",
            ),
        ),
    ]



"""Tests for the admin interface"""

from django.urls import reverse
from django.utils import timezone

import pytest
import requests
from pyquery import PyQuery

from log_outgoing_requests.models import OutgoingRequestsLog, OutgoingRequestsLogConfig


#
# test display
#
@pytest.mark.django_db
def test_decoded_content_display(admin_client):
    """Assert that decoded request/response bodies are properly displayed"""

    log = OutgoingRequestsLog.objects.create(
        id=1,
        req_body=b"I'm a lumberjack and I'm okay.",
        res_body=b"I sleep all night and work all day.",
        timestamp=timezone.now(),
    )
    url = reverse(
        "admin:log_outgoing_requests_outgoingrequestslog_change", args=(log.pk,)
    )

    response = admin_client.get(url)
    assert response.status_code == 200

    html = response.content.decode("utf-8")
    doc = PyQuery(html)
    request_body = doc.find(".request-body-decoded").text()
    response_body = doc.find(".response-body-decoded").text()

    assert request_body == "I'm a lumberjack and I'm okay."
    assert response_body == "I sleep all night and work all day."


#
# test override of settings
#
@pytest.mark.django_db
def test_admin_override_db_save(requests_mock, request_mock_kwargs):
    """Assert that saving logs can be disabled in admin"""

    config = OutgoingRequestsLogConfig.get_solo()
    config.save_to_db = "no"
    config.save()

    requests_mock.get(**request_mock_kwargs)
    requests.get(
        request_mock_kwargs["url"], headers=request_mock_kwargs["request_headers"]
    )

    request_log = OutgoingRequestsLog.objects.last()

    assert request_log is None


@pytest.mark.django_db
def test_admin_override_save_body(requests_mock, request_mock_kwargs):
    """Assert that saving body can be disabled in admin"""

    config = OutgoingRequestsLogConfig.get_solo()
    config.save_body = "no"
    config.save()

    requests_mock.get(**request_mock_kwargs)
    requests.get(
        request_mock_kwargs["url"], headers=request_mock_kwargs["request_headers"]
    )

    request_log = OutgoingRequestsLog.objects.last()

    assert request_log.res_body == b""


@pytest.mark.django_db
def test_admin_override_max_content_length(requests_mock, request_mock_kwargs):
    """Assert that `max_content_length` can be overriden in admin"""

    config = OutgoingRequestsLogConfig.get_solo()
    config.max_content_length = "10"
    config.save()

    requests_mock.get(**request_mock_kwargs)
    requests.get(
        request_mock_kwargs["url"], headers=request_mock_kwargs["request_headers"]
    )

    request_log = OutgoingRequestsLog.objects.last()

    assert request_log.res_body == b""



from django import forms
from django.conf import settings
from django.contrib import admin
from django.utils.translation import gettext as _

from solo.admin import SingletonModelAdmin

from .models import OutgoingRequestsLog, OutgoingRequestsLogConfig


@admin.register(OutgoingRequestsLog)
class OutgoingRequestsLogAdmin(admin.ModelAdmin):
    fields = (
        "url",
        "hostname",
        "query_params",
        "params",
        "status_code",
        "method",
        "response_ms",
        "timestamp",
        "req_content_type",
        "res_content_type",
        "req_headers",
        "res_headers",
        "trace",
    )
    readonly_fields = fields
    list_display = (
        "hostname",
        "url",
        "params",
        "status_code",
        "method",
        "response_ms",
        "timestamp",
    )
    list_filter = ("method", "timestamp", "status_code", "hostname")
    search_fields = ("url", "params", "hostname")
    date_hierarchy = "timestamp"
    show_full_result_count = False
    change_form_template = "log_outgoing_requests/change_form.html"

    class Media:
        css = {
            "all": ("log_outgoing_requests/css/admin.css",),
        }

    def has_add_permission(self, request):
        return False

    def has_change_permission(self, request, obj=None):
        return False

    @admin.display(description=_("Query parameters"))
    def query_params(self, obj):
        return obj.query_params


class ConfigAdminForm(forms.ModelForm):
    class Meta:
        model = OutgoingRequestsLogConfig
        fields = "__all__"
        help_texts = {
            "save_to_db": _(
                "Whether request logs should be saved to the database (default: {default})."
            ).format(default=settings.LOG_OUTGOING_REQUESTS_DB_SAVE),
            "save_body": _(
                "Whether the body of the request and response should be logged (default: "
                "{default})."
            ).format(default=settings.LOG_OUTGOING_REQUESTS_DB_SAVE_BODY),
        }


@admin.register(OutgoingRequestsLogConfig)
class OutgoingRequestsLogConfigAdmin(SingletonModelAdmin):
    form = ConfigAdminForm



# NOTE: Avoid import Django specifics at the module level to prevent circular imports.
# The handler is loaded eagerly at django startup when configuring settings.
import logging
import traceback
from datetime import datetime
from logging import LogRecord
from typing import Union, cast
from urllib.parse import urlparse

from requests.models import PreparedRequest, Response


class RequestLogRecord(LogRecord):
    requested_at: datetime
    req: PreparedRequest
    res: Response


AnyLogRecord = Union[LogRecord, RequestLogRecord]


def is_request_log_record(record: AnyLogRecord) -> bool:
    attrs = ("requested_at", "req", "res")
    if any(not hasattr(record, attr) for attr in attrs):
        return False
    return True


class DatabaseOutgoingRequestsHandler(logging.Handler):
    def emit(self, record: AnyLogRecord):
        from .models import OutgoingRequestsLog, OutgoingRequestsLogConfig
        from .utils import process_body

        config = cast(OutgoingRequestsLogConfig, OutgoingRequestsLogConfig.get_solo())
        if not config.save_logs_enabled:
            return

        # skip requests not coming from the library requests
        if not record or not is_request_log_record(record):
            return
        # Typescript type predicates would be cool here :)
        record = cast(RequestLogRecord, record)

        scrubbed_req_headers = record.req.headers.copy()

        if "Authorization" in scrubbed_req_headers:
            scrubbed_req_headers["Authorization"] = "***hidden***"

        trace = traceback.format_exc() if record.exc_info else ""

        parsed_url = urlparse(record.req.url)
        kwargs = {
            "url": record.req.url,
            "hostname": parsed_url.netloc,
            "params": parsed_url.params,
            "status_code": record.res.status_code,
            "method": record.req.method,
            "timestamp": record.requested_at,
            "response_ms": int(record.res.elapsed.total_seconds() * 1000),
            "req_headers": self.format_headers(scrubbed_req_headers),
            "res_headers": self.format_headers(record.res.headers),
            "trace": trace,
        }

        if config.save_body_enabled:
            # check request
            processed_request_body = process_body(record.req, config)
            if processed_request_body.allow_saving_to_db:
                kwargs.update(
                    {
                        "req_content_type": processed_request_body.content_type,
                        "req_body": processed_request_body.content,
                        "req_body_encoding": processed_request_body.encoding,
                    }
                )

            # check response
            processed_response_body = process_body(record.res, config)
            if processed_response_body.allow_saving_to_db:
                kwargs.update(
                    {
                        "res_content_type": processed_response_body.content_type,
                        "res_body": processed_response_body.content,
                        "res_body_encoding": processed_response_body.encoding,
                    }
                )

        OutgoingRequestsLog.objects.create(**kwargs)

    def format_headers(self, headers):
        return "\n".join(f"{k}: {v}" for k, v in headers.items())



"""Integration tests for the core functionality of the library"""

import logging

import pytest
import requests
from freezegun import freeze_time

from log_outgoing_requests.datastructures import ContentType
from log_outgoing_requests.models import OutgoingRequestsLog


#
# Local pytest fixtures
#
@pytest.fixture()
def request_variants(requests_mock):
    return [
        ("GET", requests.get, requests_mock.get),
        ("POST", requests.post, requests_mock.post),
        ("PUT", requests.put, requests_mock.put),
        ("PATCH", requests.patch, requests_mock.patch),
        ("DELETE", requests.delete, requests_mock.delete),
        ("HEAD", requests.head, requests_mock.head),
    ]


@pytest.fixture()
def expected_headers():
    return (
        f"User-Agent: python-requests/{requests.__version__}\n"
        "Accept-Encoding: gzip, deflate\n"
        "Accept: */*\n"
        "Connection: keep-alive\n"
        "Authorization: ***hidden***\n"
        "Content-Type: application/json\n"
        "Content-Length: 24"
    )


#
# Tests
#
@pytest.mark.django_db
def test_data_is_logged(requests_mock, request_mock_kwargs, caplog):
    with caplog.at_level(logging.DEBUG):
        requests_mock.get(**request_mock_kwargs)
        requests.get(
            request_mock_kwargs["url"], headers=request_mock_kwargs["request_headers"]
        )

    records = caplog.records
    assert records[1].levelname == "DEBUG"
    assert records[1].name == "requests"
    assert records[1].msg == "Outgoing request"


@pytest.mark.django_db
@freeze_time("2021-10-18 13:00:00")
def test_data_is_saved(request_mock_kwargs, request_variants, expected_headers):
    for method, request_func, request_mock in request_variants:
        request_mock(**request_mock_kwargs)
        response = request_func(
            request_mock_kwargs["url"],
            headers=request_mock_kwargs["request_headers"],
            json={"test": "request data"},
        )

        assert response.status_code == 200

        request_log = OutgoingRequestsLog.objects.last()

        assert request_log.method == method
        assert request_log.hostname == "example.com:8000"
        assert request_log.params == ""
        assert request_log.query_params == "version=2.0"
        assert request_log.response_ms == 0
        assert request_log.trace == ""
        assert str(request_log) == "example.com:8000 at 2021-10-18 13:00:00+00:00"
        assert (
            request_log.timestamp.strftime("%Y-%m-%d %H:%M:%S") == "2021-10-18 13:00:00"
        )
        # headers
        assert request_log.req_headers == expected_headers
        assert (
            request_log.res_headers == "Date: Tue, 21 Mar 2023 15:24:08 GMT\n"
            "Content-Type: application/json\nContent-Length: 25"
        )
        # request body
        assert request_log.req_content_type == "application/json"
        assert bytes(request_log.req_body) == b'{"test": "request data"}'
        assert request_log.req_body_encoding == "utf-8"
        # response body
        assert request_log.res_content_type == "application/json"
        assert bytes(request_log.res_body) == b'{"test": "response data"}'
        assert request_log.res_body_encoding == "utf-8"


#
# test decoding of binary content
#
@pytest.mark.parametrize(
    "content, encoding, expected",
    [
        (b"test\x03\xff\xff{\x03}", "utf-8", "test\x03��{\x03}"),
        (b"test\x03\xff\xff{\x03}", "utx-99", "test\x03��{\x03}"),
        (b"test{\x03\xff\xff\x00d", "", "test{\x03��\x00d"),
    ],
)
@pytest.mark.django_db
def test_decoding_of_binary_content(
    content, encoding, expected, requests_mock, request_mock_kwargs_binary, settings
):
    """
    Assert that decoding of binary contents works with:
        - correct encoding
        - wrong (e.g. misspelled) encoding
        - missing encoding
    """
    settings.LOG_OUTGOING_REQUESTS_CONTENT_TYPES = [
        ContentType(pattern="binary", default_encoding=encoding)
    ]

    request_mock_kwargs_binary["content"] = content

    requests_mock.post(**request_mock_kwargs_binary)
    response = requests.post(
        request_mock_kwargs_binary["url"],
        headers=request_mock_kwargs_binary["request_headers"],
        data=content,
    )

    assert response.status_code == 200

    request_log = OutgoingRequestsLog.objects.last()

    assert request_log.response_body_decoded == expected
    assert request_log.request_body_decoded == expected


@pytest.mark.django_db
def test_authorization_header_is_hidden(requests_mock, request_mock_kwargs):
    requests_mock.get(**request_mock_kwargs)
    requests.get(
        request_mock_kwargs["url"], headers=request_mock_kwargs["request_headers"]
    )

    log = OutgoingRequestsLog.objects.get()

    assert "Authorization: ***hidden***" in log.req_headers


@pytest.mark.django_db
def test_disable_save_db(request_mock_kwargs, request_variants, caplog, settings):
    """Assert that data is logged but not saved to DB when setting is disabled"""

    settings.LOG_OUTGOING_REQUESTS_DB_SAVE = False

    for method, request_func, request_mock in request_variants:
        with caplog.at_level(logging.DEBUG):
            request_mock(**request_mock_kwargs)
            response = request_func(
                request_mock_kwargs["url"],
                headers=request_mock_kwargs["request_headers"],
                json={"test": "request data"},
            )

            assert response.status_code == 200

            # data is logged
            records = caplog.records
            assert records[1].levelname == "DEBUG"
            assert records[1].name == "requests"
            assert records[1].msg == "Outgoing request"

            # data is not saved
            assert OutgoingRequestsLog.objects.exists() is False


@pytest.mark.django_db
def test_disable_save_body(request_mock_kwargs, request_variants, settings):
    """Assert that request/response bodies are not saved when setting is disabled"""

    settings.LOG_OUTGOING_REQUESTS_DB_SAVE_BODY = False

    for method, request_func, request_mock in request_variants:
        request_mock(**request_mock_kwargs)
        response = request_func(
            request_mock_kwargs["url"],
            headers=request_mock_kwargs["request_headers"],
            json={"test": "request data"},
        )

        assert response.status_code == 200

        request_log = OutgoingRequestsLog.objects.last()

        assert bytes(request_log.req_body) == b""
        assert bytes(request_log.res_body) == b""


@pytest.mark.django_db
def test_content_type_not_allowed(request_mock_kwargs, request_variants, settings):
    """Assert that request/response bodies are not saved when content type is not allowed"""

    settings.LOG_OUTGOING_REQUESTS_CONTENT_TYPES = [
        ContentType(pattern="text/*", default_encoding="utf-8")
    ]

    for method, request_func, request_mock in request_variants:
        request_mock(**request_mock_kwargs)
        response = request_func(
            request_mock_kwargs["url"],
            headers=request_mock_kwargs["request_headers"],
            json={"test": "request data"},
        )

        assert response.status_code == 200

        request_log = OutgoingRequestsLog.objects.last()

        assert bytes(request_log.req_body) == b""
        assert bytes(request_log.res_body) == b""


@pytest.mark.django_db
def test_content_length_exceeded(request_mock_kwargs, request_variants, settings):
    """Assert that body is not saved when content-length exceeds pre-defined max"""

    settings.LOG_OUTGOING_REQUESTS_MAX_CONTENT_LENGTH = 10

    for method, request_func, request_mock in request_variants:
        request_mock(**request_mock_kwargs)
        response = request_func(
            request_mock_kwargs["url"],
            headers=request_mock_kwargs["request_headers"],
            json={"test": "request data"},
        )

        assert response.status_code == 200

        request_log = OutgoingRequestsLog.objects.last()

        assert bytes(request_log.res_body) == b""



import logging
from typing import Union
from urllib.parse import urlparse

from django.conf import settings
from django.core.validators import MinValueValidator
from django.db import models
from django.utils.functional import cached_property
from django.utils.translation import gettext_lazy as _

from solo.models import SingletonModel  # type: ignore

from .constants import SaveLogsChoice

logger = logging.getLogger(__name__)


class OutgoingRequestsLog(models.Model):
    url = models.URLField(
        verbose_name=_("URL"),
        max_length=1000,
        default="",
        help_text=_("The url of the outgoing request."),
    )

    # hostname is added so we can filter on it in the admin page
    hostname = models.CharField(
        verbose_name=_("Hostname"),
        max_length=255,
        default="",
        help_text=_("The netloc/hostname part of the url."),
    )
    params = models.TextField(
        verbose_name=_("Parameters"),
        blank=True,
        help_text=_("The parameters (if they exist)."),
    )
    status_code = models.PositiveIntegerField(
        verbose_name=_("Status code"),
        null=True,
        blank=True,
        help_text=_("The status code of the response."),
    )
    method = models.CharField(
        verbose_name=_("Method"),
        max_length=10,
        blank=True,
        help_text=_("The type of request method."),
    )
    req_content_type = models.CharField(
        verbose_name=_("Request content type"),
        max_length=50,
        default="",
        help_text=_("The content type of the request."),
    )
    res_content_type = models.CharField(
        verbose_name=_("Response content type"),
        max_length=50,
        default="",
        help_text=_("The content type of the response."),
    )
    req_headers = models.TextField(
        verbose_name=_("Request headers"),
        default="",
        help_text=_("The request headers."),
    )
    res_headers = models.TextField(
        verbose_name=_("Response headers"),
        default="",
        help_text=_("The response headers."),
    )
    req_body = models.BinaryField(
        verbose_name=_("Request body"),
        default=b"",
        help_text=_("The request body."),
    )
    res_body = models.BinaryField(
        verbose_name=_("Response body"),
        default=b"",
        help_text=_("The response body."),
    )
    req_body_encoding = models.CharField(
        max_length=24,
        default="",
    )
    res_body_encoding = models.CharField(
        max_length=24,
        default="",
    )
    response_ms = models.PositiveIntegerField(
        verbose_name=_("Response in ms"),
        default=0,
        help_text=_("This is the response time in ms."),
    )
    timestamp = models.DateTimeField(
        verbose_name=_("Timestamp"),
        help_text=_("This is the date and time the API call was made."),
    )
    trace = models.TextField(
        verbose_name=_("Trace"),
        default="",
        help_text=_("Text providing information in case of request failure."),
    )

    class Meta:
        verbose_name = _("Outgoing Requests Log")
        verbose_name_plural = _("Outgoing Requests Logs")

    def __str__(self):
        return ("{hostname} at {date}").format(
            hostname=self.hostname, date=self.timestamp
        )

    @cached_property
    def url_parsed(self):
        return urlparse(self.url)

    @property
    def query_params(self):
        return self.url_parsed.query

    def _decode_body(self, content: Union[bytes, memoryview], encoding: str) -> str:
        """
        Decode body for use in template.

        If the stored encoding is not found (either because it is empty or because of
        spelling errors etc.), we decode "blindly", replacing chars that could not be
        decoded.

        Inspired on :meth:`requests.models.Response.text`, which is Apache 2.0 licensed.
        """
        try:
            return str(content, encoding, errors="replace")
        except LookupError:
            # A LookupError is raised if the encoding was not found which could
            # indicate a misspelling or similar mistake.
            return str(content, errors="replace")

    @cached_property
    def request_body_decoded(self) -> str:
        """
        Decoded request body for use in template.
        """
        return self._decode_body(self.req_body, self.req_body_encoding)

    @cached_property
    def response_body_decoded(self) -> str:
        """
        Decoded response body for use in template.
        """
        return self._decode_body(self.res_body, self.res_body_encoding)


def get_default_max_content_length():
    """
    Get default value for max content length from settings.
    """
    return settings.LOG_OUTGOING_REQUESTS_MAX_CONTENT_LENGTH


class OutgoingRequestsLogConfig(SingletonModel):
    """Configuration options for request logging."""

    save_to_db = models.CharField(
        _("Save logs to database"),
        max_length=11,
        choices=SaveLogsChoice.choices,
        default=SaveLogsChoice.use_default,
    )
    save_body = models.CharField(
        _("Save request + response body"),
        max_length=11,
        choices=SaveLogsChoice.choices,
        default=SaveLogsChoice.use_default,
    )
    max_content_length = models.IntegerField(
        _("Maximal content size"),
        validators=[MinValueValidator(0)],
        default=get_default_max_content_length,
        help_text=_(
            "The maximal size of the request/response content (in bytes). "
            "If 'Require content length' is not checked, this setting has no effect."
        ),
    )

    @property
    def save_logs_enabled(self):
        """
        Use configuration option or settings to determine if logs should be saved.
        """
        if self.save_to_db == SaveLogsChoice.use_default:
            return settings.LOG_OUTGOING_REQUESTS_DB_SAVE
        return self.save_to_db == SaveLogsChoice.yes

    @property
    def save_body_enabled(self):
        """
        Use configuration option or settings to determine if log bodies should be saved.
        """
        if self.save_body == SaveLogsChoice.use_default:
            return settings.LOG_OUTGOING_REQUESTS_DB_SAVE_BODY
        return self.save_body == SaveLogsChoice.yes

    class Meta:
        verbose_name = _("Outgoing Requests Log Configuration")



import django

# Taken from djangorestframework, see
# https://github.com/encode/django-rest-framework/blob/376a5cbbba3f8df9c9db8c03a7c8fa2a6e6c05f4/rest_framework/compat.py#LL156C1-L177C10
#
# License:
#
# Copyright © 2011-present, [Encode OSS Ltd](https://www.encode.io/).
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

if django.VERSION >= (4, 2):
    # Django 4.2+: use the stock parse_header_parameters function
    # Note: Django 4.1 also has an implementation of parse_header_parameters
    #       which is slightly different from the one in 4.2, it needs
    #       the compatibility shim as well.
    from django.utils.http import parse_header_parameters  # type: ignore
else:
    # Django <= 4.1: create a compatibility shim for parse_header_parameters
    from django.http.multipartparser import parse_header

    def parse_header_parameters(line):
        # parse_header works with bytes, but parse_header_parameters
        # works with strings. Call encode to convert the line to bytes.
        main_value_pair, params = parse_header(line.encode())
        return main_value_pair, {
            # parse_header will convert *some* values to string.
            # parse_header_parameters converts *all* values to string.
            # Make sure all values are converted by calling decode on
            # any remaining non-string values.
            k: v if isinstance(v, str) else v.decode()
            for k, v in params.items()
        }


__all__ = ["parse_header_parameters"]



# Generated by Django 3.2.15 on 2023-03-29 16:25

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="OutgoingRequestsLog",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "url",
                    models.URLField(
                        blank=True,
                        default="",
                        help_text="The url of the outgoing request.",
                        max_length=1000,
                        verbose_name="URL",
                    ),
                ),
                (
                    "hostname",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="The hostname part of the url.",
                        max_length=255,
                        verbose_name="Hostname",
                    ),
                ),
                (
                    "params",
                    models.TextField(
                        blank=True,
                        help_text="The parameters (if they exist).",
                        verbose_name="Parameters",
                    ),
                ),
                (
                    "status_code",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="The status code of the response.",
                        null=True,
                        verbose_name="Status code",
                    ),
                ),
                (
                    "method",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="The type of request method.",
                        max_length=10,
                        verbose_name="Method",
                    ),
                ),
                (
                    "req_content_type",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="The content type of the request.",
                        max_length=50,
                        verbose_name="Request content type",
                    ),
                ),
                (
                    "res_content_type",
                    models.CharField(
                        blank=True,
                        default="",
                        help_text="The content type of the response.",
                        max_length=50,
                        verbose_name="Response content type",
                    ),
                ),
                (
                    "req_headers",
                    models.TextField(
                        blank=True,
                        help_text="The request headers.",
                        null=True,
                        verbose_name="Request headers",
                    ),
                ),
                (
                    "res_headers",
                    models.TextField(
                        blank=True,
                        help_text="The response headers.",
                        null=True,
                        verbose_name="Response headers",
                    ),
                ),
                (
                    "response_ms",
                    models.PositiveIntegerField(
                        blank=True,
                        default=0,
                        help_text="This is the response time in ms.",
                        verbose_name="Response in ms",
                    ),
                ),
                (
                    "timestamp",
                    models.DateTimeField(
                        help_text="This is the date and time the API call was made.",
                        verbose_name="Timestamp",
                    ),
                ),
                (
                    "trace",
                    models.TextField(
                        blank=True,
                        help_text="Text providing information in case of request failure.",
                        null=True,
                        verbose_name="Trace",
                    ),
                ),
            ],
            options={
                "verbose_name": "Outgoing Requests Log",
                "verbose_name_plural": "Outgoing Requests Logs",
            },
        ),
    ]
